/* $OpenLDAP: pkg/ldap/libraries/libldap_r/rdwr.c,v 1.12 1999/12/08 22:54:33 kdz Exp $ */
/*
** This is an improved implementation of Reader/Writer locks does
** not protect writers from starvation.  That is, if a writer is
** currently waiting on a reader, any new reader will get
** the lock before the writer.
**
** Does not support cancellation nor does any status checking.
*/

/********************************************************
 * Adapted from:
 *	"Programming with Posix Threads"
 *		by David R Butenhof
 *		Addison-Wesley 
 ********************************************************
 */

#include "port_threads_portable.h"

#include <ac/stdlib.h>

#include <ac/errno.h>

#include <assert.h>

#include "port_pvt_thread.h"

int 
port_pvt_thread_rdwr_init( port_pvt_thread_rdwr_t *rw )
{
	assert( rw != NULL );

	memset( rw, 0, sizeof(port_pvt_thread_rdwr_t) );

	/* we should check return results */
	port_pvt_thread_mutex_init( &rw->ltrw_mutex );
	port_pvt_thread_cond_init( &rw->ltrw_read , PORT_THREAD_CONDATTR_DEFAULT);
	port_pvt_thread_cond_init( &rw->ltrw_write, PORT_THREAD_CONDATTR_DEFAULT);

	rw->ltrw_valid = PORT_PVT_THREAD_RDWR_VALID;
	return 0;
}

int 
port_pvt_thread_rdwr_destroy( port_pvt_thread_rdwr_t *rw )
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	if( rw->ltrw_valid != PORT_PVT_THREAD_RDWR_VALID )
		return PORT_PVT_THREAD_EINVAL;

	port_pvt_thread_mutex_lock( &rw->ltrw_mutex );

	/* active threads? */
	if( rw->ltrw_r_active > 0 || rw->ltrw_w_active > 0) {
		port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );
		return PORT_PVT_THREAD_EBUSY;
	}

	/* waiting threads? */
	if( rw->ltrw_r_wait > 0 || rw->ltrw_w_wait > 0) {
		port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );
		return PORT_PVT_THREAD_EBUSY;
	}

	rw->ltrw_valid = 0;

	port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );

	port_pvt_thread_mutex_destroy( &rw->ltrw_mutex );
	port_pvt_thread_cond_destroy( &rw->ltrw_read );
	port_pvt_thread_cond_destroy( &rw->ltrw_write );

	return 0;
}

int port_pvt_thread_rdwr_rlock( port_pvt_thread_rdwr_t *rw )
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	if( rw->ltrw_valid != PORT_PVT_THREAD_RDWR_VALID )
		return PORT_PVT_THREAD_EINVAL;

	port_pvt_thread_mutex_lock( &rw->ltrw_mutex );

	if( rw->ltrw_w_active > 0 ) {
		/* writer is active */

		rw->ltrw_r_wait++;

		do {
			port_pvt_thread_cond_wait(
				&rw->ltrw_read, &rw->ltrw_mutex );
		} while( rw->ltrw_w_active > 0 );

		rw->ltrw_r_wait--;
	}

	rw->ltrw_r_active++;

	port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );

	return 0;
}

int port_pvt_thread_rdwr_rtrylock( port_pvt_thread_rdwr_t *rw )
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	if( rw->ltrw_valid != PORT_PVT_THREAD_RDWR_VALID )
		return PORT_PVT_THREAD_EINVAL;

	port_pvt_thread_mutex_lock( &rw->ltrw_mutex );

	if( rw->ltrw_w_active > 0) {
		port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );
		return PORT_PVT_THREAD_EBUSY;
	}

	rw->ltrw_r_active++;

	port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );

	return 0;
}

int port_pvt_thread_rdwr_runlock( port_pvt_thread_rdwr_t *rw )
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	if( rw->ltrw_valid != PORT_PVT_THREAD_RDWR_VALID )
		return PORT_PVT_THREAD_EINVAL;

	port_pvt_thread_mutex_lock( &rw->ltrw_mutex );

	rw->ltrw_r_active--;

	if (rw->ltrw_r_active == 0 && rw->ltrw_w_wait > 0 ) {
		port_pvt_thread_cond_signal( &rw->ltrw_write );
	}

	port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );

	return 0;
}

int port_pvt_thread_rdwr_wlock( port_pvt_thread_rdwr_t *rw )
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	if( rw->ltrw_valid != PORT_PVT_THREAD_RDWR_VALID )
		return PORT_PVT_THREAD_EINVAL;

	port_pvt_thread_mutex_lock( &rw->ltrw_mutex );

	if ( rw->ltrw_w_active > 0 || rw->ltrw_r_active > 0 ) {
		rw->ltrw_w_wait++;

		do {
			port_pvt_thread_cond_wait(
				&rw->ltrw_write, &rw->ltrw_mutex );
		} while ( rw->ltrw_w_active > 0 || rw->ltrw_r_active > 0 );

		rw->ltrw_w_wait--;
	}

	rw->ltrw_w_active++;

	port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );

	return 0;
}

int port_pvt_thread_rdwr_wtrylock( port_pvt_thread_rdwr_t *rw )
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	if( rw->ltrw_valid != PORT_PVT_THREAD_RDWR_VALID )
		return PORT_PVT_THREAD_EINVAL;

	port_pvt_thread_mutex_lock( &rw->ltrw_mutex );

	if ( rw->ltrw_w_active > 0 || rw->ltrw_r_active > 0 ) {
		port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );
		return PORT_PVT_THREAD_EBUSY;
	}

	rw->ltrw_w_active++;

	port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );

	return 0;
}

int port_pvt_thread_rdwr_wunlock( port_pvt_thread_rdwr_t *rw )
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	if( rw->ltrw_valid != PORT_PVT_THREAD_RDWR_VALID )
		return PORT_PVT_THREAD_EINVAL;

	port_pvt_thread_mutex_lock( &rw->ltrw_mutex );

	rw->ltrw_w_active--;

	if (rw->ltrw_r_wait > 0) {
		port_pvt_thread_cond_broadcast( &rw->ltrw_read );

	} else if (rw->ltrw_w_wait > 0) {
		port_pvt_thread_cond_signal( &rw->ltrw_write );
	}

	port_pvt_thread_mutex_unlock( &rw->ltrw_mutex );

	return 0;
}

#if defined(PORT_THREADS_DEBUG)

/* just for testing, 
 * return 0 if false, suitable for assert(port_pvt_thread_rdwr_Xchk(rdwr))
 * 
 * Currently they don't check if the calling thread is the one 
 * that has the lock, just that there is a reader or writer.
 *
 * Basically sufficent for testing that places that should have
 * a lock are caught.
 */

int port_pvt_thread_rdwr_readers(port_pvt_thread_rdwr_t *rw)
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	return( rw->ltrw_r_active );
}

int port_pvt_thread_rdwr_writers(port_pvt_thread_rdwr_t *rw)
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	return( rw->ltrw_w_active );
}

int port_pvt_thread_rdwr_active(port_pvt_thread_rdwr_t *rw)
{
	assert( rw != NULL );
	assert( rw->ltrw_valid == PORT_PVT_THREAD_RDWR_VALID );

	return(port_pvt_thread_rdwr_readers(rw) +
	       port_pvt_thread_rdwr_writers(rw));
}

#endif /* PORT_THREADS_DEBUG */
